// a simple class stack - feel free to use or copy this code

#ifndef stack_H
#define stack_H

// #define ASSERT_YES

#include <stdlib.h>     // for NULL
#include <assert.h>     // for macro assert()

template <class T>
class stack
{
   typedef unsigned long ulong;

public:
   #ifndef ASSERT_YES
      // exceptions
      class MemoryOut {};
      class StackEmpty {};
   #endif

   stack( ): sp(NULL) {}
   stack(const stack& s) { stackcpy(s); }
   ~stack( ) { clear( ); }
   const stack& operator=(const stack&);

   void push(const T& val);
   T pop( );

   void clear( );

protected:
   struct node
   {
      T val;
      node* prev;
   };
   node* sp;        // stack pointer

   void stackcpy(const stack&);

   // this is and must be inline
   // note that it is NOT a member function
   // (this is what static means)
   inline static node* new_node() // inline here is optional
   {
      node* pnode=new node;

      #ifndef ASSERT_YES
         if (pnode==NULL) throw MemoryOut( );
      #else
         assert(pnode != NULL);
      #endif

      return pnode;
   }
};

template <class T>
void stack<T>::stackcpy(const stack& s)
{
   n=s.n;

   if (s.sp != NULL)
   {
      node* tail= new_node();

      const node* ssp=s.sp; // cursor on the list s.sp
      node* tmp;

      // first copy the top
      sp= tail;
      sp->val = ssp->val;

      // then all the other elements in order
      while ((ssp = ssp->prev) != NULL)
      {
         tmp=new_node();

         tail->prev = tmp;
         tail = tmp;
         tail->val = ssp->val;
      }
      tail->prev = NULL;
   }
   else
      sp=NULL;
}

template <class T>
const stack<T>& stack<T>::operator=(const stack& s)
{
   clear();
   stackcpy(s);
   return (*this);
}

template <class T>
void stack<T>::push(const T& val)
{
   node* tmp= new_node();

   tmp->val=val;
   tmp->prev=sp;
   sp=tmp;
}

template <class T>
T stack<T>::pop( )
{
   #ifndef ASSERT_YES
      if (sp==NULL) throw StackEmpty( );
   #else
      assert(sp != NULL);
   #endif

   T val= sp->val;
   node* tmp= sp;
   sp=sp->prev;
   delete tmp;

   return(val);
}

template <class T>
void stack<T>::clear( )
{
   for (node* q = sp; q != NULL; q = sp)        // or sp != NULL
   {
      sp = sp->prev;
      delete q;
   }
}

#endif
