                                // stack.h

// a simple class stack - feel free to use or copy this code
// FINAL VERSION

#ifndef stack_H
#define stack_H 

// choose from: ASSERT_YES (default), NOERROR_TRAP, EXCEPTION
#define EXCEPTION

#include <stdlib.h>     // for NULL
#include <assert.h>     // for macro assert()

template <class T>
class stack
{
   typedef unsigned long ulong;

public:
   #ifdef EXCEPTION
      // exceptions
      class MemoryOut {};
      class StackEmpty {};
   #endif

   stack( ): sp(NULL), n(0) {}
   stack(const stack& s) { stackcpy(s); }
   ~stack( ) { clear( ); }
   const stack& operator=(const stack&);

   void push(const T& val);
   T pop( );
   T& top( );
   T& dtop( );

   ulong size( ) { return n; }
   bool empty( ) { return sp==NULL; }
   bool full( );
   void clear( );

   bool operator == (const stack&);
   bool operator != (const stack& s) { return !((*this)==s); }

protected:
   struct node
   {
      T val;
      node* prev;
   };

   node* sp;    // stack pointer
   ulong n;     // number of elements

   void stackcpy(const stack&);

   inline static node* new_node() // inline here is optional
   {
      node* pnode=new node;

      #ifdef NOERROR_TRAP
         // don't check heap overflow!
      #elif defined(EXCEPTION)
         if (pnode==NULL) throw MemoryOut( );
      #else // ASSERT_YES
         assert(pnode != NULL);        
      #endif

      return pnode;
   }

   void ExceptionIfEmpty()
   {
      #ifdef EXCEPTION
         if (sp==NULL) throw StackEmpty( );
      #elif defined(NOERROR_TRAP)
         // don't trap this error!
      #else // ASSERT_YES
         assert(sp != NULL);
      #endif
   }
};

#endif
